"""
Visual Stimuli -- still very alpha
"""

WIN = None

from autopilot import prefs

import threading
import os
import sys
import datetime
if sys.version_info >= (3,0):
    from queue import Queue, Empty
else:
    from queue import Queue, Empty

#print(prefs._PREFS.items())
if prefs.get( 'CONFIG'):
    if 'VISUAL' in prefs.get('CONFIG'):
        from psychopy import visual, core
else:
    Warning('No CONFIG attr set in prefs, dont know if youre set up for visual stim. not importing psychopy')


class Visual(object):
    """Metaclass for visual stimuli"""
    callback = None


    def __init__(self, debug=False):
        # psychopy Window
        self.win = None
        self.duration = None
        self.ppo = None # psychopy object
        #self.get_window()

        self.debug = debug

        self.clock = core.Clock()
        self.draw_time = 0

        self.thread_lock = threading.Lock()

    def get_window(self):
        """
        Try to get the global visuals.WIN window,
        if it hasn't been made yet, make it.
        """
        if globals()['WIN'] is not None:
            self.win = globals()['WIN']
        else:
            try:
                global WIN
                WIN = visual.Window(winType="glfw", color=(-1., -1., -1.), fullscr=True, size=(1280,720))
                self.win = WIN
            except:
                Exception("Couldn't get psychopy Window!")




class Grating(Visual):
    """
    Moving grating
    """

    PARAMS = ['angle', 'freq', 'rate', 'phase',
              'mask', 'pos', 'size', 'duration']

    def __init__(self, angle, freq, rate, phase=0,
                 mask="gauss", pos=(0., 0.), size=(2,2),
                 duration=5000., debug=False):
        super(Grating, self).__init__(debug=debug)

        self.angle = angle
        self.freq = freq
        self.rate = rate
        self.phase = phase
        self.mask = mask
        self.pos = pos
        self.size = size
        self.duration = duration

        self.play_evt = threading.Event()
        self.stop_evt = threading.Event()
        self.stop_evt.clear()
        self.q = Queue()

        self.threadfn()

    def threadfn(self):
        self.thread = threading.Thread(target=self._thread)
        self.thread.start()

    def _thread(self):
        self.get_window()
        self.clock = core.Clock()
        self.draw_time = 0

        # init psychopy object
        self.ppo = visual.GratingStim(
            self.win,
            mask=self.mask,
            pos=self.pos,
            size=self.size,
            sf=self.freq,
            ori=self.angle,
            phase=self.phase)

        while not self.stop_evt.is_set():
            self.play_evt.wait()

            if self.debug:
                self.win.recordFrameIntervals = True

            # reset stim
            self.ppo.phase = self.phase

            start_time = self.clock.getTime()
            end_time = start_time + (self.duration / 1000.0)
            while self.clock.getTime() < end_time:
                try:
                    attrchange = self.q.get_nowait()
                except Empty:
                    attrchange = None
                if attrchange is not None:
                    if attrchange[0] == 'shift':
                        self.ppo.ori = self.ppo.ori + attrchange[1]

                self.update()
                self.ppo.draw()
                self.win.flip()

            # another flip clears the screen
            self.win.flip()
            self.play_evt.clear()

            if self.debug:
                path = os.path.join(prefs.get('DATADIR'), 'frameintervals_'+datetime.datetime.now().isoformat()+'.csv')
                self.win.saveFrameIntervals(path)
                self.win.recordFrameIntervals = False




    def set(self, attr, value):
        """
        Set psychopy attrs

        Args:
            attr ():
            value ():

        Returns:

        """
        attr_map = {
            'mask':'mask', 'pos':'pos', 'size':'size',
            'freq':'sf', 'angle':'ori', 'phase':'phase'
        }

        if attr in ('mask', 'pos', 'size', 'freq', 'angle', 'phase'):
            self.q.put_nowait((attr_map[attr], value))



    def update(self):
        """advance the psychopy object one frame"""

        # get change since last draw, divide by rate
        dt = self.clock.getTime()-self.draw_time
        self.ppo.phase = (self.ppo.phase + self.rate*dt) % 1.0
        self.draw_time = self.clock.getTime()


    def play(self, attr, val):
        self.q.put((attr, val))
        self.play_evt.set()



class Grating_Continuous(Grating):
    """
    Moving grating that goes continuously, but changes angles
    """

    PARAMS = ['angle', 'freq', 'rate', 'phase',
              'mask', 'pos', 'size']

    def __init__(self, **kwargs):
        super(Grating_Continuous, self).__init__(**kwargs)
        self.stop_flag = threading.Event()
        self.stop_flag.clear()

    def start(self):
        self.stop_flag.set()
        while self.stop_flag:
            self.play()
        self.win.flip()

    def stop(self):
        self.stop_flag.clear()


    def play(self):
        """
        draws continuously
        Returns:

        """
        self.update()
        self.ppo.draw()
        self.win.flip()










import numpy as np
from scipy.spatial import distance

from autopilot.transform.transforms import Transform


class Distance(Transform):
    """
    Given an n_samples x n_dimensions array, compute pairwise or mean distances
    """
    format_in = {'type': np.ndarray}
    format_out = {'type': np.ndarray}

    def __init__(self,
                 pairwise: bool=False,
                 n_dim: int = 2,
                 metric: str='euclidean',
                 squareform: bool=True,
                 *args, **kwargs):
        """

        Args:
            pairwise (bool): If False (default), return mean distance. if True, return all distances
            n_dim (int): number of dimensions (input array will be filtered like ``input[:,0:n_dim]``
            metric (str): any metric acceptable to :func:`scipy.spatial.distance.pdist
            squareform (bool): if pairwise is True, if True return square distance matrix, otherwise return compressed distance matrix (dist(X[i], X[j] = y[i*j])
            *args:
            **kwargs:
        """

        super(Distance, self).__init__(*args, **kwargs)

        self.pairwise = pairwise
        self.n_dim = n_dim
        self.metric = metric
        self.squareform = squareform

    def process(self, input: np.ndarray):

        # filter to input_dimension
        input = input[:,0:self.n_dim]

        output = distance.pdist(input, metric=self.metric)

        if self.pairwise:
            if self.squareform:
                output = distance.squareform(output)
        else:
            output = np.mean(output)

        return output


class Angle(Transform):
    """
    Get angle between line formed by two points and horizontal axis
    """

    format_in = {'type': np.ndarray}
    format_out = {'type': float}

    def __init__(self, abs=True, degrees=True, *args, **kwargs):
        super(Angle, self).__init__(*args, **kwargs)
        self.abs = abs
        self.degrees = degrees

    def process(self, input):
        angle = np.arctan2(input[1][1]-input[0][1], input[1][0]-input[0][0])
        if self.abs:
            angle += np.pi
        if self.degrees:
            angle = angle*(180/np.pi)
        return angle


"""
This is a scrappy first draft of a stimulus manager that will be built out
to incorporate arbitrary stimulus logic. For now you can subclass `Stim_Manager` and
redefine `next_stim`

TODO:
    Make this more general, for more than just sounds.
"""

import os
import pdb
from collections import deque
import numpy as np
from autopilot import prefs
if prefs.get('AGENT').upper() == 'PILOT':
    if 'AUDIO' in prefs.get('CONFIG') or prefs.get('AUDIOSERVER') is not None:
        from autopilot.stim.sound import sounds
        # TODO be loud about trying to init sounds when not in config

def init_manager(stim):
    if 'manager' in stim.keys():
        if stim['manager'] in MANAGER_MAP.keys():
            manager = stim['manager']
            return MANAGER_MAP[manager](stim)
        else:
            Exception('Couldnt find stim manager of type: {}'.format(stim['type']))
    else:
        return Stim_Manager(stim)


class Stim_Manager(object):
    """
    Yield sounds according to some set of rules.

    Currently implemented:

    * correction trials - If a subject continually answers to one side incorrectly, keep
        the correct answer on the other side until they answer in that direction
    * bias correction - above some bias threshold, skew the correct answers to the less-responded side

    Attributes:
        stimuli (dict): Dictionary of instantiated stimuli like::

            {'L': [Tone1, Tone2, ...], 'R': [Tone3, Tone4, ...]}

        target ('L', 'R'): What is the correct port?
        distractor ('L', 'R'): What is the incorrect port?
        response ('L', 'R'): What was the last response?
        correct (0, 1): Was the last response correct?
        last_stim: What was the last stim? (one of `self.stimuli`)
        correction (bool): Are we doing correction trials?
        correction_trial (bool): Is this a correction trial?
        last_was_correction (bool): Was the last trial a correction trial?
        correction_pct (float): proportion of trials that are correction trials
        bias: False, or a bias correction mode.

    """

    # Metaclass for managing stimuli...
    def __init__(self, stim=None):
        """
        Args:
            stim (dict): Dictionary describing sound stimuli, in a format like::

                {
                'L': [{'type':'tone',...},{...}],
                'R': [{'type':'tone',...},{...}]
                }
        """
        self.stimuli = {}

        self.target = None  # What is the correct port?
        self.distractor = None  # What is the incorrect port
        self.response = None  # What was the last response?
        self.correct = 0  # Was the last response correct?
        self.last_stim = None  # What was the last stim?

        # Correction trials

        self.correction = False  # Are we doing correction trials
        self.correction_trial = False  # Is this a correction trial?
        self.last_was_correction = False  # Was the last trial a correction trial?
        self.correction_pct = 0.1  # proportion of trials that are correction trials

        # Bias correction
        self.bias = False  # or a bias correction mode

        # if we're being init'd as a superclass, no stim passed.
        if stim:

            # if we're doing sounds init them here
            # TODO: Make SoundManger subclass
            if 'sounds' in stim.keys():
                self.init_sounds(stim['sounds'])

    def do_correction(self, correction_pct = 0.5):
        """
        Called to set correction trials to True and correction percent.

        Args:
            correction_pct (float): Proportion of trials that should randomly be set to be correction trials.
        """
        self.correction = True
        self.correction_pct = correction_pct

    def do_bias(self, **kwargs):
        """
        Instantiate a :class:`.Bias_Correction` module

        Args:
            kwargs: parameters to initialize :class:`.Bias_Correction` with.
        """
        self.bias = Bias_Correction(**kwargs)

    def init_sounds(self, sound_dict):
        """
        Instantiate sound objects, using the 'type' value to choose an object from
        :data:`.sounds.SOUND_LIST` .

        Args:
            sound_dict (dict): a dictionary like::
                {
                'L': [{'type':'tone',...},{...}],
                'R': [{'type':'tone',...},{...}]
                }
        """
        # sounds should be
        # Iterate through sounds and load them to memory
        for k, v in sound_dict.items():
            # If multiple sounds on one side, v will be a list
            if isinstance(v, list):
                self.stimuli[k] = []
                for sound in v:
                    # We send the dict 'sound' to the function specified by 'type' and '
                    # ' as kwargs
                    self.stimuli[k].append(sounds.SOUND_LIST[sound['type']](**sound))
            # If not a list, a single sound
            else:
                self.stimuli[k] = [sounds.SOUND_LIST[v['type']](**v)]

    def set_triggers(self, trig_fn):
        """
        Give a callback function to all of our stimuli for when the stimulus ends.

        Note:
            Stimuli need a `set_trigger` method.

        Args:
            trig_fn (callable): A function to be given to stimuli via `set_trigger`
        """
        # set a callback function for when the stimulus ends
        for k, v in self.stimuli.items():
            for astim in v:
                astim.set_trigger(trig_fn)

    def make_punishment(self, type, duration):
        """
        Warning:
            Not Implemented

        Args:
            type:
            duration:
        """
        # types: timeout, noise
        # If we want a punishment sound...
        # if self.punish_sound:
        #     self.stimuli['punish'] = sounds.Noise(self.punish_dur)
        #     #change_to_green = lambda: self.hardware['LEDS']['C'].set_color([0, 255, 0])
        #     #self.stimuli['punish'].set_trigger(change_to_green)
        pass

    def play_punishment(self):
        """
        Warning:
            Not Implemented
        """
        pass

    def next_stim(self):
        """
        Compute and return the next stimulus

        If we are doing correction trials, compute that.

        Same thing with bias correction.

        Otherwise, randomly select a stimulus to present.

        Returns:
            ('L'/'R' Target, 'L'/'R' distractor, Stimulus to present)
        """
        # compute and return the next stim

        # first: if we're doing correction trials, compute that
        if self.correction:
            self.correction_trial = self.compute_correction()
            if self.correction_trial:
                return self.target, self.distractor, self.last_stim

        # otherwise we check for bias correction
        # it will return a threshold for random choice
        if self.bias:
            threshold = self.bias.next_bias()
        else:
            threshold = 0.5

        if np.random.rand()<threshold:
            self.target = 'L'
        else:
            self.target = 'R'

        if self.target == 'L':
            self.distractor = 'R'
        elif self.target == 'R':
            self.distractor = 'L'

        self.last_stim = np.random.choice(self.stimuli[self.target])

        return self.target, self.distractor, self.last_stim

    def compute_correction(self):
        """
        If `self.correction` is true, compute correction trial logic during
        `next_stim`.

        * If the last trial was a correction trial and the response to it wasn't correct, return True
        * If the last trial was a correction trial and the response was correct, return False
        * If the last trial as not a correction trial, but a randomly generated float is less than `correction_pct`, return True.

        Returns:
            bool: whether this trial should be a correction trial.

        """
        # if we are doing a correction trial this time return true

        # if this is the first trial, we can't do correction trials now can we
        if self.target is None:
            return False

        # if the last trial was a correction trial and we didn't get it correct,
        # then this is a correction trial too
        if (self.correction_trial or self.last_was_correction) and not self.correct:
            return True
        # if the last trial was a correction trial and we just corrected, no correction trial this time
        elif (self.correction_trial or self.last_was_correction) and self.correct:
            self.last_was_correction = False
            return False
        # if last trial was not a correction trial we spin  *to test* for one
        elif np.random.rand() < self.correction_pct:
            self.last_was_correction = True
            return False
        else:
            return False

    def update(self, response, correct):
        """
        At the end of a trial, update the status of our internal variables
        with the outcome of the trial.

        Args:
            response ('L', 'R'): How the subject responded
            correct (0, 1): Whether the response was correct.
        """
        self.response = response
        self.correct = correct
        if self.bias:
            self.bias.update(response, self.target)

    def end(self):
        """
        End all of our stim. Stim should have an `.end()` method of their own

        """

        for side, v in self.stimuli.items():
            for stim in v:
                try:
                    stim.end()
                except AttributeError:
                    print('stim does not have an end method! \n{}'.format(str(stim)))

class Proportional(Stim_Manager):
    """
    Present groups of stimuli with a particular frequency.

    Frequencies do not need to add up to 1, groups will be selected with the frequency
    (frequency)/(sum(frequencies)).

    Arguments:
        stim (dict): Dictionary with the structure::

            {'manager': 'proportional',
             'type': 'sounds',
             'groups': (
                 {'name':'group_name',
                  'frequency': 0.2,
                  'sounds':{
                      'L': [{Tone1_params}, {Tone2_params}...],
                      'R': [{Tone3_params}, {Tone4_params}...]
                  }
                },
                {'name':'second_group',
                  'frequency': 0.8,
                  'sounds':{
                      'L': [{Tone1_params}, {Tone2_params}...],
                      'R': [{Tone3_params}, {Tone4_params}...]
                  }
                })
            }

    Attributes:
        stimuli (dict): A dictionary of stimuli organized into groups
        groups (dict): A dictionary mapping group names to frequencies

    """

    def __init__(self, stim):
        super(Proportional, self).__init__()

        self.stimuli = {}

        self.frequency_type = None

        if stim['type'] == 'sounds':
            if 'groups' in stim.keys():
                self.frequency_type = "within_groups"
                # top-level groups, choose group then choose side
                self.init_sounds_grouped(stim['groups'])
                self.store_groups(stim)
            else:
                self.frequency_type = "within_side"
                # second-level frequencies, side is chosen and then
                # probability from within a side
                self.init_sounds_individual(stim['sounds'])



    def init_sounds_grouped(self, sound_stim):
        """
        Instantiate sound objects similarly to :class:`.Stim_Manager`, just organizes them into groups.

        Args:
            sound_stim (tuple, list): an iterator like::
                (
                 {'name':'group_name',
                  'frequency': 0.2,
                  'sounds': {
                      'L': [{Tone1_params}, {Tone2_params}...],
                      'R': [{Tone3_params}, {Tone4_params}...]
                    }
                },
                {'name':'second_group',
                  'frequency': 0.8,
                  'sounds':{
                      'L': [{Tone1_params}, {Tone2_params}...],
                      'R': [{Tone3_params}, {Tone4_params}...]
                  }
                })
        """
        # Iterate through sounds and load them to memory
        self.stimuli = {}

        if isinstance(sound_stim, tuple) or isinstance(sound_stim, list):
            for group in sound_stim:
                group_name = group['name']

                # instantiate sounds
                self.stimuli[group_name] = {}
                for k, v in group['sounds'].items():
                    if isinstance(v, list):
                        self.stimuli[group_name][k] = []
                        for sound in v:
                            # We send the dict 'sound' to the function specified by 'type' and '
                            # ' as kwargs
                            self.stimuli[group_name][k].append(sounds.SOUND_LIST[sound['type']](**sound))
                    # If not a list, a single sound
                    else:
                        self.stimuli[group_name][k] = [sounds.SOUND_LIST[v['type']](**v)]


    def init_sounds_individual(self, sound_stim):
        """
        Initialize sounds with individually set presentation frequencies.

        .. todo::

            This method reflects the need for managers to have a unified schema,
            which will be built in a future release of Autopilot.

        Args:
            sound_stim (dict): Dictionary of {'side':[sound_params]} to generate sound stimuli

        Returns:

        """
        self.stim_freqs = {}
        for side, sound_params in sound_stim.items():
            self.stimuli[side] = []
            self.stim_freqs[side] = []
            if isinstance(sound_params, list):
                for sound in sound_params:
                    self.stimuli[side].append(sounds.SOUND_LIST[sound['type']](**sound))
                    self.stim_freqs[side].append(float(sound['management']['frequency']))
            else:
                self.stimuli[side].append(sounds.SOUND_LIST[sound_params['type']](**sound_params))
                self.stim_freqs[side].append(float(sound_params['management']['frequency']))

        # normalize frequencies within sides to sum to 1
        for side, freqs in self.stim_freqs.items():
            side_sum = np.sum(freqs)
            self.stim_freqs[side] = tuple([float(f)/side_sum for f in freqs])

    def store_groups(self, stim):
        """
        store groups and frequencies
        """
        self.groups = {}
        # also store (later) as tuples for choice b/c dicts are unordered
        group_names = []
        group_freqs = []
        for group in stim['groups']:
            group_name = group['name']
            frequency = float(group['frequency'])
            self.groups[group_name] = frequency
            group_names.append(group_name)
            group_freqs.append(frequency)

        self.group_names = tuple(group_names)
        group_freqs = np.array(group_freqs).astype(np.float)
        group_freqs = group_freqs/np.sum(group_freqs)

        self.group_freqs = tuple(group_freqs)



    def set_triggers(self, trig_fn):
        """
        Give a callback function to all of our stimuli for when the stimulus ends.

        Note:
            Stimuli need a `set_trigger` method.

        Args:
            trig_fn (callable): A function to be given to stimuli via `set_trigger`
        """
        # set a callback function for when the stimulus ends

        if self.frequency_type == "within_group":
            for _, group in self.stimuli.items():
                    for _, v in group.items():
                        for astim in v:
                            astim.set_trigger(trig_fn)
        elif self.frequency_type == "within_side":
            for side, sound_group in self.stimuli.items():
                for astim in sound_group:
                    astim.set_trigger(trig_fn)

        else:
            ValueError('Dont know how to set triggers')


    def next_stim(self):
        """
        Compute and return the next stimulus

        If we are doing correction trials, compute that.

        Same thing with bias correction.

        Otherwise, randomly select a stimulus to present, weighted by its group frequency.

        Returns:
            ('L'/'R' Target, 'L'/'R' distractor, Stimulus to present)
        """
        # compute and return the next stim

        # first: if we're doing correction trials, compute that
        if self.correction:
            self.correction_trial = self.compute_correction()
            if self.correction_trial:
                return self.target, self.distractor, self.last_stim

        # otherwise we check for bias correction
        # it will return a threshold for random choice
        if self.bias:
            threshold = self.bias.next_bias()
        else:
            threshold = 0.5

        # choose side
        if np.random.rand()<threshold:
            self.target = 'L'
        else:
            self.target = 'R'

        if self.target == 'L':
            self.distractor = 'R'
        elif self.target == 'R':
            self.distractor = 'L'

        if self.frequency_type == "within_group":

            # pick a stimulus based on group frequency
            group = np.random.choice(self.group_names, p=self.group_freqs)
            # within that group pick a random stimulus
            self.last_stim = np.random.choice(self.stimuli[group][self.target])

        elif self.frequency_type == "within_side":
            self.last_stim = np.random.choice(self.stimuli[self.target],
                                              p=self.stim_freqs[self.target])
        else:
            ValueError('Dont know what freq type we are')

        return self.target, self.distractor, self.last_stim





class Bias_Correction(object):
    """
    Basic Bias correction module. Modifies the threshold of random stimulus
    choice based on history of biased responses.

    Attributes:
        responses (:class:`collections.deque`): History of prior responses
        targets (:class:`collections.deque`): History of prior targets.
    """

    def __init__(self, mode='thresholded_linear', thresh=.2, window=100):
        """
        Args:
            mode: One of the following:

                * `'thresholded linear'` : above some threshold, do linear bias correction
                    eg. if response rate 65% left, make correct be right 65% of the time

            thresh (float): threshold above chance, ie. 0.2 means has to be 70% biased in window
            window (int): number of trials to calculate bias over
        """
        self.mode = mode
        self.threshold = float(thresh)
        self.window = int(window)
        self.responses = deque(maxlen=self.window)
        self.targets = deque(maxlen=self.window)

    def next_bias(self):
        """
        Compute the next bias depending on `self.mode`

        Returns:
            float: Some threshold :class:`.Stim_Manager` uses to decide left vs right.

        """

        # compute the bias threshold
        if self.mode == 'thresholded_linear':
            return self.thresholded_linear()

    def thresholded_linear(self):
        """
        If we are above the threshold, linearly correct the rate of
        presentation to favor the rarely responded side.

        eg. if response rate 65% left, make correct be right 65% of the time

        Returns:
            float: 0.5-bias, where bias is the difference between the mean response and mean target.
        """
        # if we are above threshold, return the bias
        bias = np.mean(self.responses)-np.mean(self.targets)
        if np.abs(bias)>self.threshold:
            return 0.5-bias
        else:
            return 0.5


    def update(self, response, target):
        """
        Store some new response and target values

        Args:
            response ('R', 'L'): Which side the subject responded to
            target ('R', 'L'): The correct side.
        """
        if isinstance(response, str):
            if response == "R":
                response = 1.0
            elif response == "L":
                response = 0.0

        if isinstance(target, str):
            if target == "R":
                target = 1.0
            elif target == "L":
                target = 0.0

        self.responses.append(float(response))
        self.targets.append(float(target))

MANAGER_MAP = {
    'proportional': Proportional
}





# class Reward_Manager(object):
#     def __init__(self):
#         pass




import datetime
import itertools
import tables
import threading
from copy import copy
import typing

import numpy as np

import autopilot.hardware.gpio
from autopilot.tasks import Task
from autopilot.stim import init_manager
from autopilot.stim.sound import sounds as stim_sounds
from autopilot.hardware import gpio
from collections import OrderedDict as odict
from autopilot.core.networking import Net_Node
from autopilot.core.utils import find_recursive

from autopilot import prefs
import pdb
import pickle

# This declaration allows Subject to identify which class in this file contains the task class. Could also be done with __init__ but yno I didnt for no reason.
# TODO: Move this to __init__
TASK = 'Nafc'

class Nafc(Task):
    """
    A Two-alternative forced choice task.

    *(can't have number as first character of class.)*

    **Stages**

    * **request** - compute stimulus, set request trigger in center port.
    * **discrim** - respond to input, set reward/punishment triggers on target/distractor ports
    * **reinforcement** - deliver reward/punishment, end trial.

    Attributes:
        target ("L", "R"): Correct response
        distractor ("L", "R"): Incorrect response
        stim : Current stimulus
        response ("L", "R"): Response to discriminand
        correct (0, 1): Current trial was correct/incorrect
        correction_trial (bool): If using correction trials, last trial was a correction trial
        trial_counter (:class:`itertools.count`): Which trial are we on?
        discrim_playing (bool): Is the stimulus playing?
        bailed (0, 1): Subject answered before stimulus was finished playing.
        current_stage (int): As each stage is reached, update for asynchronous event reference

    """
    STAGE_NAMES = ["request", "discrim", "reinforcement"]

    # Class attributes


    # List of needed params, returned data and data format.
    # Params are [name]={'tag': Human Readable Tag, 'type': 'int', 'float', 'bool', etc.}
    PARAMS = odict()
    # TODO: Reward no longer just duration -- fix with parameter structure
    PARAMS['reward']         = {'tag':'Reward Duration (ms)',
                                'type':'int'}
    PARAMS['req_reward']     = {'tag':'Request Rewards',
                                'type':'bool'}
    PARAMS['punish_stim']   = {'tag':'White Noise Punishment',
                                'type':'bool'}
    PARAMS['punish_dur']     = {'tag':'Punishment Duration (ms)',
                                'type':'int'}
    PARAMS['correction']     = {'tag':'Correction Trials',
                                'type':'bool'}
    PARAMS['correction_pct'] = {'tag':'% Correction Trials',
                                'type':'int',
                                'depends':{'correction':True}}
    PARAMS['bias_mode']      = {'tag':'Bias Correction Mode',
                                'type':'list',
                                'values':{'None':0, 'Proportional':1, 'Thresholded Proportional':2}}
    PARAMS['bias_threshold'] = {'tag': 'Bias Correction Threshold (%)',
                                'type':'int',
                                'depends':{'bias_mode':2}}
    #PARAMS['timeout']        = {'tag':'Delay Timeout (ms)',
    #                            'type':'int'}
    PARAMS['stim']           = {'tag':'Sounds',
                                'type':'sounds'}

    # Set plot params, which data should be plotted, its default shape, etc.
    # TODO: Plots should take the default type, but options panel should be able to set - eg. corrects are done by rolling mean as default, but can be made points
    PLOT = {
        'data': {
            'target'   : 'point',
            'response' : 'segment',
            'correct'  : 'rollmean'
        },
        'chance_bar'  : True, # Draw a red bar at 50%
        'roll_window' : 50 # number of trials to roll window over
    }

    # PyTables Data descriptor
    # for numpy data types see http://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html#arrays-dtypes-constructing
    class TrialData(tables.IsDescription):
        # This class allows the Subject object to make a data table with the correct data types. You must update it for any new data you'd like to store
        trial_num = tables.Int32Col()
        target = tables.StringCol(1)
        response = tables.StringCol(1)
        correct = tables.Int32Col()
        correction = tables.Int32Col()
        RQ_timestamp = tables.StringCol(26)
        DC_timestamp = tables.StringCol(26)
        bailed = tables.Int32Col()

    HARDWARE = {
        'POKES':{
            'L': gpio.Digital_In,
            'C': gpio.Digital_In,
            'R': gpio.Digital_In
        },
        'LEDS':{
            # TODO: use LEDs, RGB vs. white LED option in init
            'L': gpio.LED_RGB,
            'C': gpio.LED_RGB,
            'R': gpio.LED_RGB
        },
        'PORTS':{
            'L': gpio.Solenoid,
            'C': gpio.Solenoid,
            'R': gpio.Solenoid
        }
    }


    def __init__(self, stage_block=None, stim=None, reward=50, req_reward=False,
                 punish_stim=False, punish_dur=100, correction=False, correction_pct=50.,
                 bias_mode=False, bias_threshold=20, current_trial=0, stim_light=True, **kwargs):
        """
        Args:
            stage_block (:class:`threading.Event`): Signal when task stages complete.
            stim (dict): Stimuli like::


                "sounds": {
                    "L": [{"type": "Tone", ...}],
                    "R": [{"type": "Tone", ...}]
                }

            reward (float): duration of solenoid open in ms
            req_reward (bool): Whether to give a water reward in the center port for requesting trials
            punish_stim (bool): Do a white noise punishment stimulus
            punish_dur (float): Duration of white noise in ms
            correction (bool): Should we do correction trials?
            correction_pct (float):  (0-1), What proportion of trials should randomly be correction trials?
            bias_mode (False, "thresholded_linear"): False, or some bias correction type (see :class:`.managers.Bias_Correction` )
            bias_threshold (float): If using a bias correction mode, what threshold should bias be corrected for?
            current_trial (int): If starting at nonzero trial number, which?
            stim_light (bool): Should the LED be turned blue while the stimulus is playing?
            **kwargs:
        """
        super(Nafc, self).__init__()

        # Fixed parameters
        # Because the current protocol is json.loads from a string,
        # we should explicitly type everything to be safe.
        if isinstance(reward, dict):
            self.reward = reward
        else:
            self.reward         = {'type':'duration',
                                   'value': float(reward)}
        self.req_reward     = bool(req_reward)
        self.punish_stim   = bool(punish_stim)
        self.punish_dur     = float(punish_dur)
        self.correction     = bool(correction)
        self.correction_pct = float(correction_pct)/100
        self.bias_mode      = bias_mode
        self.bias_threshold = float(bias_threshold)/100
        self.stim_light      = bool(stim_light)
        #self.timeout        = int(timeout)

        # Variable Parameters
        self.target = None
        self.distractor = None
        self.stim = None
        self.response = None
        self.correct = None
        self.correction_trial = False
        self.trial_counter = itertools.count(int(current_trial))
        self.current_trial = int(current_trial)
        #self.discrim_finished = False # Set to true once the discrim stim has finished, used for punishing leaving C early
        self.discrim_playing = False
        self.current_stage = None # Keep track of stages so some asynchronous callbacks know when it's their turn
        self.bailed = 0

        # We make a list of the variables that need to be reset each trial so it's easier to do so
        self.resetting_variables = [self.response, self.bailed]

        # This allows us to cycle through the task by just repeatedly calling self.stages.next()
        stage_list = [self.request, self.discrim, self.reinforcement]
        self.num_stages = len(stage_list)
        self.stages = itertools.cycle(stage_list)

        # Initialize hardware
        self.init_hardware()
        self.logger.debug('Hardware initialized')

        # Set reward values for solenoids
        # TODO: Super inelegant, implement better with reward manager
        if self.reward['type'] == "volume":
            self.set_reward(vol=self.reward['value'])
        else:
            self.set_reward(duration=self.reward['value'])

        # Initialize stim manager
        if not stim:
            raise RuntimeError("Cant instantiate task without stimuli!")
        else:
            self.stim_manager = init_manager(stim)

        # give the sounds a function to call when they end
        self.stim_manager.set_triggers(self.stim_end)

        if self.correction:
            self.stim_manager.do_correction(self.correction_pct)

        if bias_mode:
            self.stim_manager.do_bias(mode=self.bias_mode,
                                      thresh=self.bias_threshold)
        self.logger.debug('Stimulus manager initialized')

        # If we aren't passed an event handler
        # (used to signal that a trigger has been tripped),
        # we should warn whoever called us that things could get a little screwy
        if not stage_block:
            raise Warning('No stage_block Event() was passed, youll need to handle stage progression on your own')
        else:
            self.stage_block = stage_block

    #
    # def center_out(self, pin, level, tick):
    #     """
    #
    #     """
    #     # Called when something leaves the center pin,
    #     # We use this to handle the subject leaving the port early
    #     if self.discrim_playing:
    #         self.bail_trial()


    ##################################################################################
    # Stage Functions
    ##################################################################################
    def request(self,*args,**kwargs):
        """
        Stage 0: compute stimulus, set request trigger in center port.

        Returns:
            data (dict): With fields::

                {
                'target': self.target,
                'trial_num' : self.current_trial,
                'correction': self.correction_trial,
                'type': stimulus type,
                **stim.PARAMS
                }

        """
        # Set the event lock
        self.stage_block.clear()

        # Reset all the variables that need to be
        for v in self.resetting_variables:
            v = None

        # reset triggers if there are any left
        self.triggers = {}

        # get next stim
        self.target, self.distractor, self.stim = self.stim_manager.next_stim()
        # buffer it
        self.stim.buffer()

        # if we're doing correction trials, check if this is one
        if self.correction:
            self.correction_trial = self.stim_manager.correction_trial

        # Set sound trigger and LEDs
        self.triggers['C'] = [self.stim.play, self.stim_start]
        if self.stim_light:
            change_to_blue = lambda: self.hardware['LEDS']['C'].set([0, 0, 255])
            self.triggers['C'].append(change_to_blue)
        else:
            # just turn the center light off and side lights on immediately.
            turn_off = lambda: self.set_leds({
                'L': [0,255,0],
                'R': [0,255,0]
            })
            #turn_off = lambda: self.hardware['LEDS']['C'].set([0,0,0])
            self.triggers['C'].append(turn_off)

        if self.req_reward:
            self.triggers['C'].append(self.hardware['PORTS']['C'].open)


        self.current_trial = next(self.trial_counter)
        data = {
            'target':self.target,
            'trial_num' : self.current_trial,
            'correction':self.correction_trial
        }
        # get stim info and add to data dict
        sound_info = {k:getattr(self.stim, k) for k in self.stim.PARAMS}
        data.update(sound_info)
        data.update({'type':self.stim.type})

        self.current_stage = 0

        # wait on punish block
        # FIXME: Only waiting to test whether this is where the bug that hangs after this stage is
        self.punish_block.wait(20)

        # set to green in the meantime
        self.set_leds({'C': [0, 255, 0]})


        return data

    def discrim(self,*args,**kwargs):
        """
        Stage 1:  respond to input, set reward/punishment triggers on target/distractor ports

        Returns:
            data (dict): With fields::
                {
                'RQ_timestamp': datetime.datetime.now().isoformat(),
                'trial_num': self.current_trial,
                }

        """
        # moust just poked in center, set response triggers
        self.stage_block.clear()

        self.triggers[self.target] = [lambda: self.respond(self.target), self.hardware['PORTS'][self.target].open]
        self.triggers[self.distractor] = [lambda: self.respond(self.distractor), self.punish]

        # TODO: Handle timeout

        # Only data is the timestamp
        data = {'RQ_timestamp': datetime.datetime.now().isoformat(),
                'trial_num': self.current_trial}
        self.current_stage = 1
        return data

    def reinforcement(self,*args,**kwargs):
        """
        Stage 2 - deliver reward/punishment, end trial.

        Returns:
            data (dict): With fields::

                 {
                'DC_timestamp': datetime.datetime.now().isoformat(),
                'response': self.response,
                'correct': self.correct,
                'bailed': self.bailed,
                'trial_num': self.current_trial,
                'TRIAL_END': True
                }
        """
        # We do NOT clear the task event flag here because we want
        # the pi to call the next stage immediately
        # We are just filling in the last data
        # and performing any calculations we need for the next trial
        if self.bailed:
            self.bailed = 0
            data = {
                'DC_timestamp': datetime.datetime.now().isoformat(),
                'bailed':1,
                'trial_num': self.current_trial,
                'TRIAL_END':True
            }
            return data

        if self.response == self.target:
            self.correct = 1
        else:
            self.correct = 0

        # update stim manager
        self.stim_manager.update(self.response, self.correct)


        data = {
            'DC_timestamp': datetime.datetime.now().isoformat(),
            'response':self.response,
            'correct':self.correct,
            'bailed':0,
            'trial_num': self.current_trial,
            'TRIAL_END':True
        }
        self.current_stage = 2
        return data

    def punish(self):
        """
        Flash lights, play punishment sound if set
        """
        # TODO: If we're not in the last stage (eg. we were timed out after stim presentation), reset stages
        self.punish_block.clear()

        if self.punish_stim:
            self.stim_manager.play_punishment()

        # self.set_leds()
        self.flash_leds()
        threading.Timer(self.punish_dur / 1000., self.punish_block.set).start()


    def respond(self, pin):
        """
        Set self.response

        Args:
            pin: Pin to set response to
        """
        self.response = pin

    def stim_start(self):
        """
        mark discrim_playing = true
        """
        self.discrim_playing = True


    def stim_end(self):
        """
        called by stimulus callback

        set outside lights blue
        """
        # Called by the discrim sound's table trigger when playback is finished
        # Used in punishing leaving early
        self.discrim_playing = False
        #if not self.bailed and self.current_stage == 1:
        if self.stim_light:
            self.set_leds({'L':[0,255,0], 'R':[0,255,0]})

    # def bail_trial(self):
    #     # If a timer ends or the subject pulls out too soon, we punish and bail
    #     self.bailed = 1
    #     self.triggers = {}
    #     self.punish()
    #     self.stage_block.set()

    # def clear_triggers(self):
    #     for pin in self.hardware.values():
    #         pin.clear_cb()

    def flash_leds(self):
        """
        flash lights for punish_dir
        """
        for k, v in self.hardware['LEDS'].items():
            if isinstance(v, gpio.LED_RGB):
                v.flash(self.punish_dur)

class Nafc_Gap(Nafc):
    PARAMS = copy(Nafc.PARAMS)
    del PARAMS['punish_stim']
    PARAMS['noise_amplitude'] = {'tag':'Amplitude of continuous white noise',
                                 'type': 'float'}

    def __init__(self, noise_amplitude = 0.01, **kwargs):
        """
        A Mild variation of :class:`Nafc` that starts continuous white noise that plays
        continuously while the task is active.

        Args:
            noise_amplitude (float): Multiplier used to scale amplitude of continuous noise
            **kwargs: passed to :class:`Nafc`
        """

        # Can't really have a white noise punishment when there is continuous noise
        kwargs['punish_stim'] = False
        kwargs['stim_light'] = False
        super(Nafc_Gap, self).__init__(**kwargs)

        self.logger.debug('starting background sound')
        self.noise_amplitude = noise_amplitude
        self.noise_duration = 10*1000 # 10 seconds
        self.noise = stim_sounds.Noise(duration=self.noise_duration,
                                  amplitude=self.noise_amplitude)

        self.noise.play_continuous()
        self.logger.debug('background sound started')


    def end(self):
        """
        Stop the task, ending the continuous white noise.
        """
        self.noise.stop_continuous()
        super(Nafc_Gap, self).end()


class Nafc_Gap_Laser(Nafc_Gap):
    PARAMS = copy(Nafc_Gap.PARAMS)
    PARAMS['laser_probability'] = {'tag': 'Probability (of trials whose targets match laser_mode) of laser being turned on (0-1)',
                                   'type':'float'}
    PARAMS['laser_mode'] = {'tag':'Laser Mode, laser will be possible when target == ?',
        'type':'list',
        'values':{
            'L':0,
            'R':1,
            'Both':2
        }}
    PARAMS['laser_freq'] = {'tag': 'Laser Pulse Frequency (Hz), list-like [20, 30]',
                            'type': 'str'}
    PARAMS['laser_duty_cycle'] = {'tag': 'Laser Duty Cycle (0-1), list-like [0.1, 0.2]',
                                  'type': 'str'}
    PARAMS['laser_durations'] = {'tag': 'Laser durations (ms), list-like [10, 20]. if blank, use durations from stimuli',
                                 'type': 'str'}
    PARAMS['arena_led_mode'] = {'tag': 'Arena LED Mode: always ON vs. on for longest stim duration during requests',
                                'type': 'list',
                                'values':{'ON': 0, 'STIM': 1}}

    HARDWARE = copy(Nafc_Gap.HARDWARE)

    HARDWARE['LASERS'] = {
        'LR': gpio.Digital_Out
    }

    HARDWARE['LEDS']['TOP'] = gpio.Digital_Out

    TrialData = copy(Nafc_Gap.TrialData)
    TrialData.laser = tables.Int32Col()
    TrialData.laser_duration = tables.Float32Col()
    TrialData.laser_freq = tables.Float32Col()
    TrialData.laser_duty_cycle = tables.Float32Col()


    def __init__(self,
                 laser_probability: float,
                 laser_mode: str,
                 laser_freq: typing.Union[str, list],
                 laser_duty_cycle: typing.Union[str, list],
                 laser_durations: typing.Union[str, list],
                 arena_led_mode: str = 'ON',
                 **kwargs):
        """
        Gap detection task with ability to control lasers via TTL logic for optogenetics

        :attr:`.laser_freq`, :attr:`.laser_duty_cycle`, and :attr:`.laser_durations` can be passed
        either as an integer (actually typically a string because of the way the value is pulled from the protocol wizard),
        or as a list -- the product of values for all three are generated and presented equiprobably
        (eg. if ``laser_freq = 20, laser_duty_cycle=[0.1, 0.2, 0.3], laser_durations = [1, 2, 4, 8]`` were passed,
        then 1*3*4=12 different laser conditions would be possible.

        .. note::

            Subclasses like these will be made obsolete with the completion of stimulus managers

        Args:
            laser_probability (float): if trial satisfies ``laser_mode``, probability that laser will be
            laser_mode ('L', 'R', or 'Both'): Selects whether the laser is to be presented when :attr:`.target` is ``'L', 'R'`` or Either.
            laser_freq (str, list): Single value or list of possible laser frequencies in Hz
            laser_duty_cycle (str, list): Single value or list of possible duty cycles from 0-1
            laser_durations (str, list): Single value or list of possible laser durations (total time laser is on) in ms
            arena_led_mode ('ON', 'STIM'): Whether the overhead LED should always be 'ON', or whether it should be illuminated for the duration of the longest stimulus at every request

        Attributes:
            laser_conditions (tuple): tuple of dicts of laser conditions, of format::

                {
                'freq': laser frequency,
                'duty_cycle': laser duty cycle,
                'duration': laser duration,
                'script_id': script ID for the series used by the laser Digital Out object,
                }
        """
        self.laser_probability = float(laser_probability)
        self.laser_mode = laser_mode
        self.arena_led_mode = arena_led_mode

        # accept them if we're given a list of values, otherwise they should be strings that are single values,
        # which are put in lists so they can be iterated over in the product iterator.
        self.laser_freq = laser_freq if isinstance(laser_freq, list) else [float(laser_freq)] # type: list
        self.laser_duty_cycle = laser_duty_cycle if isinstance(laser_duty_cycle, list) else [float(laser_duty_cycle)] # type: list
        self.laser_durations = laser_durations if isinstance(laser_durations, list) else [float(laser_durations)] # type: list

        self.laser_conditions = tuple() # type: typing.Tuple[typing.Dict]

        super(Nafc_Gap_Laser, self).__init__(**kwargs)

        self.init_lasers()

        # -----------------------------------
        # create a pulse for the LED that's equal to the longest stimulus duration
        # use find_recursive to find all durations
        # FIXME: implement stimulus managers properly, including API to get attributes of stimuli
        if self.arena_led_mode == "ON":
            self.hardware['LEDS']['TOP'].turn(True)
        elif self.arena_led_mode == "STIM":
            stim_durations = list(find_recursive('duration', kwargs['stim']))
            max_duration = np.max(stim_durations)
            self.hardware['LEDS']['TOP'].store_series('on', values=1, durations=max_duration )
        else:
            raise ValueError(f'arena_led_mode must be one of ON or STIM, got {self.arena_led_mode}')

    def init_lasers(self):
        """
        Given :attr:`.laser_freq`, :attr:`.laser_duty_cycle`, :attr:`.laser_durations` ,
        create series with :meth:`.Digital_Out.store_series` and populate :attr:`.laser_conditions`
        """

        # TODO: This really should be something that Digital_Out should be capable of doing -- specifying series from these params...

        # --------------------------------------
        # create description of laser pulses
        # iterate over laser condition lists,
        # create lists of values (on/off) and durations (ms)
        # use them to create pigpio scripts using the Digital_Out.store_series() method
        # --------------------------------------------------
        self.logger.debug('Creating laser and LED series')
        # create iterator
        condition_iter = itertools.product(self.laser_durations, self.laser_freq, self.laser_duty_cycle)

        conditions = []
        for duration, freq, duty_cycle in condition_iter:
            # get the durations of on and off for a single cycle
            cycle_duration = (1/freq)*1000 # convert Hz to ms
            duty_cycle_on = duty_cycle * cycle_duration
            duty_cycle_off = cycle_duration - duty_cycle_on

            # get number of repeats to make
            n_cycles = int(np.floor(duration/cycle_duration))
            durations = [duty_cycle_on, duty_cycle_off]*n_cycles
            values = [1, 0]*n_cycles

            # pad any incomplete cycles
            dur_remaining = duration-(cycle_duration*n_cycles)
            if dur_remaining < duty_cycle_on:
                durations.append(dur_remaining)
                values.append(1)
            else:
                durations.extend([duty_cycle_on, dur_remaining-duty_cycle_on])
                values.extend([1, 0])

            # create ID from params
            script_id = f"{duration}_{freq}_{duty_cycle}"


            # store pulses as pigpio scripts
            self.hardware['LASERS']['LR'].store_series(script_id, values=values, durations=durations)

            conditions.append({
                'freq':freq,
                'duty_cycle': duty_cycle,
                'duration': duration,
                'script_id': script_id
            })

        self.laser_conditions = tuple(conditions)

        self.logger.debug(f'Laser series created with {len(self.laser_conditions)} conditions')




    def request(self,*args,**kwargs):
        """
        Call the superclass request method, and then compute laser presentation logic.

        If :attr:`.target` == :attr:`.laser_mode`, spin for a laser trial depending on :attr:`.laser_probability`.

        If we present a laser on this trial, we randomly draw from :attr:`.laser_conditions` and call the appropriate script.
        """
        # call the super method
        data = super(Nafc_Gap_Laser, self).request(*args, **kwargs)

        # handle laser logic
        # if the laser_mode is fulfilled, roll for a laser
        test_laser = False
        if self.laser_mode == "L" and self.target == "L":
            test_laser = True
        elif self.laser_mode == "R" and self.target == "R":
            test_laser = True
        elif self.laser_mode == "Both":
            test_laser = True

        duration = 0
        duty_cycle = 0
        frequency = 0
        do_laser = False
        if test_laser:
            # if we've rolled correctly for a laser...
            if np.random.rand() <= self.laser_probability:
                do_laser = True

                # If we're doing laser, we don't do the stim, so we pop the first two triggers
                del self.triggers['C'][:2]

                # pick a random duration
                condition = np.random.choice(self.laser_conditions)
                duration = condition['duration']
                duty_cycle = condition['duty_cycle']
                frequency = condition['freq']
                # insert the laser triggers before the rest of the triggers
                self.triggers['C'].insert(0, lambda: self.hardware['LASERS']['LR'].series(id=condition['script_id']))

        # always turn the light on
        if self.arena_led_mode == "STIM":
            self.triggers['C'].insert(0, lambda: self.hardware['LEDS']['TOP'].series(id='on'))


        # store the data about the laser status
        data['laser'] = do_laser
        data['laser_duration'] = duration
        data['laser_duty_cycle'] = duty_cycle
        data['laser_frequency'] = frequency

        # return the data created by the original task
        return data

    def set_leds(self, color_dict=None):
        """
        Set the color of all LEDs at once.

        Override base method to exclude TOP led

        Args:
            color_dict (dict): If None, turn LEDs off, otherwise like:

                {'pin': [R,G,B],
                'pin2: [R,G,B]}


        """
        # We are passed a dict of ['pin']:[R, G, B] to set multiple colors
        # All others are turned off
        if not color_dict:
            color_dict = {}
        for k, v in self.hardware['LEDS'].items():
            if k == "TOP":
                continue
            if k in color_dict.keys():
                v.set(color_dict[k])
            else:
                v.set(0)









#
# class Nafc_Wheel(Nafc):
#     """
#     2afc using a wheel run on a child pi as the input device
#     """
#     HARDWARE = {
#         'POKES': {
#             'C': hardware.Beambreak,
#         },
#         'FLAGS': {
#             'L': hardware.Flag,
#             'R': hardware.Flag
#         },
#         'LEDS': {
#             # TODO: use LEDs, RGB vs. white LED option in init
#             'L': hardware.LED_RGB,
#             'C': hardware.LED_RGB,
#             'R': hardware.LED_RGB
#         }
#     }
#
#     PLOT = {
#         'data': {
#             'x':'shaded',
#             'response':'segment'
#         },
#         'continuous' : True
#     }
#
#     PARAMS = Nafc.PARAMS
#
#
#
#     def __init__(self, **kwargs):
#         self.init_networking(kwargs)
#
#         super(Nafc_Wheel, self).__init__(**kwargs)
#
#
#         # TODO: Update PARAMS with wheel params
#     def init_networking(self, kwargs):
#
#         self.node = Net_Node(id="T_{}".format(prefs.get('NAME')),
#                              upstream=prefs.get('NAME'),
#                              port=prefs.get('MSGPORT'),
#                              listens = {},
#                              instance=True)
#
#         value = {
#             'child': {'parent':prefs.get('NAME'), 'subject':kwargs['subject']},
#             'task_type': 'Wheel Child',
#             'subject': kwargs['subject']
#         }
#
#         self.node.send(key='CHILD', value=value)
#




#
# class Gap_2AFC(Nafc):
#     def __init__(self, **kwargs):
#
#         """
#         Args:
#             **kwargs:
#         """
#         super(Gap_2AFC, self).__init__(**kwargs)
#
#
#     def load_sounds(self):
#         # TODO: Definitely put this in a metaclass
#
#         # Iterate through sounds and load them to memory
#         for k, v in self.soundict.items():
#             # If multiple sounds on one side, v will be a list
#             if isinstance(v, list):
#                 self.sounds[k] = []
#                 for sound in v:
#                     if float(sound['duration']) == 0:
#                         self.sounds[k].append(None)
#                         continue
#                         # a zero duration gap doesn't change the continuous noise object
#
#                     # We send the dict 'sound' to the function specified by 'type' and 'SOUND_LIST' as kwargs
#                     self.sounds[k].append(sounds.SOUND_LIST[sound['type']](**sound))
#                     # Then give the sound a callback to mark when it's finished
#                     #self.sounds[k][-1].set_trigger(self.stim_end)
#             # If not a list, a single sound
#             else:
#                 if v['duration'] == 0:
#                     self.sounds[k] = [None]
#                     continue
#
#                 self.sounds[k] = sounds.SOUND_LIST[v['type']](**v)
#                 #self.sounds[k].set_trigger(self.stim_end)
#
#     def blank_trigger(self):
#         print('blank trig')
#         sys.stdout.flush()
#         #pass
#
#     def stim_end(self):
#         # Called by the discrim sound's table trigger when playback is finished
#         # Used in punishing leaving early
#
#         self.set_leds({'L':[0,255,0], 'R':[0,255,0]})
#
#     def request(self,*args,**kwargs):
#         """
#         Args:
#             args:
#             kwargs:
#         """
#         # Set the event lock
#         self.stage_block.clear()
#
#         # Reset all the variables that need to be
#         for v in self.resetting_variables:
#             v = None
#
#         self.triggers = {}
#
#
#
#         # Attempt to identify target sound
#         #TODO: Implement sound ID's better
#         #try:
#         #    self.target_sound_id = self.target_sound.id
#         #except AttributeError:
#         #    warnings.warn("Sound ID not defined! Sounds cannot be uniquely identified!")
#         #    self.target_sound_id = None
#
#         # Set sound trigger and LEDs
#         # We make two triggers to play the sound and change the light color
#
#         # set triggers
#         if self.target_sound is None:
#             sound_trigger = self.blank_trigger
#         else:
#             sound_trigger = self.target_sound.play
#
#
#         if self.req_reward is True:
#             self.triggers['C'] = [self.hardware['PORTS']['C'].open, sound_trigger, self.stim_end]
#         else:
#             self.triggers['C'] = [sound_trigger, self.stim_end]
#         self.set_leds({'C': [0, 255, 0]})
#
#         self.current_trial = self.trial_counter.next()
#         data = {
#             'target':self.target,
#             #'target_sound_id':self.target_sound_id,
#             'RQ_timestamp':datetime.datetime.now().isoformat(),
#             'trial_num' : self.current_trial
#             #'correction':self.correction_trial
#         }
#         # get sound info and add to data dict
#         if self.target_sound is None:
#             sound_info = {'duration':0, 'amplitude':0.01}
#         else:
#             sound_info = {k:getattr(self.target_sound, k) for k in self.target_sound.PARAMS}
#             data.update({'type': self.target_sound.type})
#
#         data.update(sound_info)
#
#
#         self.current_stage = 0
#         return data
#
#     def end(self):
#         for k, v in self.sounds.items():
#             if isinstance(v, list):
#                 for sound in v:
#                     try:
#                         sound.stop()
#                     except:
#                         pass
#             else:
#                 try:
#                     v.stop()
#                 except:
#                     pass
#
#         for k, v in self.hardware.items():
#             for pin, obj in v.items():
#                 if k == "LEDS":
#                     obj.set_color([0,0,0])
#                 obj.release()












"""
Hardware that uses USB
"""
import sys
import threading
import time
if sys.version_info >= (3, 0):
    from queue import Queue, Empty
else:
    from Queue import Queue, Empty

import numpy as np
from inputs import devices

from autopilot import prefs
from autopilot.core.networking import Net_Node
from autopilot.hardware import Hardware


class Wheel(Hardware):
    """
    A continuously measured mouse wheel.

    Uses a USB computer mouse.

    Warning:
        'vel' thresh_type not implemented
    """

    input   = True
    type    = "Wheel"
    trigger = False # even though this is a triggerable option, typically don't want to assign a cb and instead us a GPIO
    # TODO: Make the standard-style trigger.
    # TODO: Make wheel movements available locally with a deque

    THRESH_TYPES = ['dist', 'x', 'y', 'vel']

    MODES = ('vel_total', 'steady', 'dist', 'timed')

    MOVE_DTYPE = [('vel', 'i4'), ('dir', 'U5'), ('timestamp', 'f8')]

    def __init__(self, mouse_idx=0, fs=10, thresh=100, thresh_type='dist', start=True,
                 digi_out = False, mode='vel_total', integrate_dur=5):
        """
        Args:
            mouse_idx (int):
            fs (int):
            thresh (int):
            thresh_type ('dist'):
            start (bool):
            digi_out (:class:`~.Digital_Out`, bool):
            mode ('vel_total'):
            integrate_dur (int):
        """

        # try to get mouse from inputs
        # TODO: More robust - specify mouse by hardware attrs
        try:
            self.mouse = devices.mice[mouse_idx]
        except IndexError:
            Warning('Could not find requested mouse with index {}\nAttempting to use mouse idx 0'.format(mouse_idx))
            self.mouse = devices.mice[0]

        # frequency of our updating
        self.fs = fs
        # time between updates
        self.update_dur = 1./float(self.fs)

        self.thresh = thresh
        # thresh type can be 'dist', 'x', 'y', or 'vel'
        if thresh_type not in self.THRESH_TYPES:
            ValueError('thresh_type must be one of {}, given {}'.format(self.THRESH_TYPES, thresh_type))
        self.thresh_type = thresh_type

        # mode can be 'vel_total', 'vel_x', 'vel_y' or 'dist' - report either velocity or distance
        # mode can also be '
        # TODO: Do two parameters - type 'vel' or 'dist' and measure 'x', 'y', 'total'z
        self.mode = mode
        # TODO: Implement this

        if self.mode == "steady":
            self.thresh_val = np.array([(0, "REL_Y", 0)], dtype=self.MOVE_DTYPE)
        else:
            self.thresh_val = 0.0

        self.integrate_dur = integrate_dur


        # event to signal quitting
        self.quit_evt = threading.Event()
        self.quit_evt.clear()
        # event to signal when to start accumulating movements to trigger
        self.measure_evt = threading.Event()
        self.measure_time = 0
        # queue to I/O mouse movements summarized at fs Hz
        self.q = Queue()
        # lock to prevent race between putting and getting
        self.qlock = threading.Lock()

        self.listens = {'MEASURE':self.l_measure,
                        'CLEAR':self.l_clear,
                        'STOP':self.l_stop}
        self.node = Net_Node('wheel_{}'.format(mouse_idx),
                             upstream=prefs.get('NAME'),
                             port=prefs.get('MSGPORT'),
                             listens=self.listens,
                             )

        # if we are being used in a child object, we send our trigger via a GPIO pin
        self.digi_out = digi_out



        self.thread = None

        if start:
            self.start()


    def start(self):
        self.thread = threading.Thread(target=self._record)
        self.thread.daemon = True
        self.thread.start()

    def _mouse(self):
        while self.quit_evt:
            events = self.mouse.read()
            self.q.put(events)

    def _record(self):
        moves = np.array([], dtype=self.MOVE_DTYPE)

        threading.Thread(target=self._mouse).start()

        last_update = time.time()

        while not self.quit_evt.is_set():

            try:
                events = self.q.get_nowait()
            except Empty:
                events = None

            if events is None:
                move = np.array([(0, "REL_Y", 0)], dtype=self.MOVE_DTYPE)
            else:
                # make a numpy record array of events with 3 fields:
                # velocity, dir(ection), timestamp (system seconds)
                move = np.array([(int(event.state), event.code, float(event.timestamp))\
                                 for event in events if event.code in ('REL_X', 'REL_Y')],
                                dtype=self.MOVE_DTYPE)
            moves = np.concatenate([moves, move])

            # If we have been told to start measuring for a trigger...
            if self.measure_evt.is_set():
                do_trigger = self.check_thresh(move)
                if do_trigger:
                    self.thresh_trig()
                    self.measure_evt.clear()
                # take the integral of velocities



            # If it's time to report velocity, do it.
            nowtime = time.time()
            if (nowtime-last_update)>self.update_dur:

                # TODO: Implement distance/position reporting
                y_vel = self.calc_move(moves, 'y')
                x_vel = self.calc_move(moves, 'x')

                self.node.send(key='CONTINUOUS', value={'x':x_vel, 'y':y_vel, 't':nowtime},
                               repeat=False)

                moves = np.array([], dtype=self.MOVE_DTYPE)

                last_update = nowtime

    def check_thresh(self, move):
        """
        Updates thresh_val and checks whether it's above/below threshold

        Args:
            move (np.array): Structured array with fields ('vel', 'dir', 'timestamp')

        Returns:

        """

        # Determine whether the threshold was surpassed
        do_trigger = False
        if self.mode == 'vel_total':
            thresh_update = self.calc_move(move)
            # If instantaneous velocity is above thresh...
            if thresh_update > self.thresh:
                do_trigger = True

        elif self.mode == 'steady':
            # If movements in the recent past are below a certain value
            # self.thresh_val should be set to a structured array by l_measure
            try:
                self.thresh_val = np.concatenate([self.thresh_val, move])
            except TypeError:
                print('THRESH_VAL:', self.thresh_val, 'MOVE:', move)
            # trim to movements in the time window
            thresh_val = self.thresh_val[self.thresh_val['timestamp'] > time.time()-self.integrate_dur]

            thresh_update = self.calc_move(thresh_val)

            if (thresh_update < self.thresh) and (self.measure_time+self.integrate_dur < time.time()):
                do_trigger = True

        elif self.mode == 'dist':
            thresh_update = self.calc_move(move)
            self.thresh_val += thresh_update

            if self.thresh_val > self.thresh:
                do_trigger = True

        else:
            Warning ("mode is not defined! mode is {}".format(self.mode))


        return do_trigger

    def calc_move(self, move, thresh_type=None):
        """
        Calculate distance move depending on type (x, y, total dist)

        Args:
            move ():
            thresh_type ():

        Returns:

        """

        if thresh_type is None:
            thresh_type = self.thresh_type

        # FIXME: rly inefficient
        # get the value of the movement depending on what we're measuring
        if thresh_type == 'x':

            distance = np.sum(move['vel'][move['dir'] == "REL_X"])
        elif thresh_type == 'y':
            distance = np.sum(move['vel'][move['dir'] == "REL_Y"])
        elif thresh_type == "dist":
            x_dist = np.sum(move['vel'][move['dir'] == "REL_X"])
            y_dist = np.sum(move['vel'][move['dir'] == "REL_Y"])
            distance = np.abs(np.sqrt(float(x_dist ** 2) + float(y_dist ** 2)))

        return distance

    def thresh_trig(self):


        if self.digi_out:
            self.digi_out.pulse()

        self.measure_evt.clear()




    def assign_cb(self, trigger_fn):
        # want to have callback write an output pin -- so callback should go back to
        # the task to write a GPIO pin.
        self.trig_fn = trigger_fn

    def l_measure(self, value):
        """
        Task has signaled that we need to start measuring movements for a trigger


        Args:
            value ():
        """

        if 'mode' in value.keys():
            if value['mode'] in self.MODES:
                self.mode = value['mode']
            else:
                Warning('incorrect mode sent: {}, needs to be one of {}'.format(value['mode'], self.MODES))

        if 'thresh' in value.keys():
            self.thresh = float(value['thresh'])

        if self.mode == "steady":
            self.thresh_val = np.array([(0, "REL_Y", 0)], dtype=self.MOVE_DTYPE)
        else:
            self.thresh_val = 0.0
        self.measure_time = time.time()

        self.measure_evt.set()

        sys.stdout.flush()

    def l_clear(self, value):
        """
        Stop measuring!

        Args:
            value ():

        Returns:

        """
        self.measure_evt.clear()

    def l_stop(self, value):
        """
        Stop measuring and clear system resources
        Args:
            value ():

        Returns:

        """

        self.measure_evt.set()
        self.release()

    def release(self):
        self.quit_evt.clear()


class Scale(Hardware):
    """
    Note:
        Not implemented, working on using a digital scale to
        make weighing faster.
    """
    MODEL={
        'stamps.com':{
            'vendor_id':0x1446,
            'product_id': 0x6a73

        }
    }
    def __init__(self, model='stamps.com', vendor_id = None, product_id = None):
        """
        Args:
            model:
            vendor_id:
            product_id:
        """
        self.vendor_id = self.MODEL[model]['vendor_id']
        self.product_id = self.MODEL[model]['product_id']

        if vendor_id:
            self.vendor_id = vendor_id
        if product_id:
            self.product_id = product_id

        # find device
        self.device = usb.core.find(idVendor=self.vendor_id,
                                    idProduct=self.product_id)
        # default configuration
        self.device.set_configuration()